import json
import re
from typing import List

import aiohttp

from pynear.constants import TGAS
from pynear.dapps.core import DappClient, NEAR
from pynear.dapps.fts import FtModel
from pynear.dapps.phone.exceptions import RequestLimitError
from pynear.dapps.phone.models import NearTrustTransaction, FtTrustTransaction
from pynear.exceptions.exceptions import FunctionCallError

_PHONE_CONTRACT_ID = "phone.herewallet.near"


class Phone(DappClient):
    """
    Client to phone.herewallet.near contract
    With this contract you can send NEAR and fungible tokens to
    phone number. Reciver will get notification with link to claim tokens.
    """

    def __init__(self, account, api_key="default"):
        """

        :param account:
        :param api_key: there is limit for hash generator requests to provide spam,
         to make unlimited requests use api key. You can get it for free by contacting team@herewallet.app
        :param network: "mainnet" or "testnet"
        """
        if account.chain_id != "mainnet":
            raise ValueError("Only mainnet is supported")
        super().__init__(account)
        self._api_key = api_key

    async def _get_phone_hex(self, phone) -> str:
        """
        To calculate hash we need call herewallet api. This is necessary to prevent creation hash <> phone table.
        :param phone:
        :return: phone hash
        """
        if phone[0] != "+":
            raise ValueError("Phone number must start with +")
        phone = re.sub(r"\D", "", phone)
        async with aiohttp.ClientSession() as session:
            r = await session.get(
                f"https://api.herewallet.app/api/v1/phone/calc_phone_hash?phone={phone}",
                headers={
                    "Content-Type": "application/json",
                    "Authorization": self._api_key,
                },
            )
            if r.status == 403:
                raise RequestLimitError(
                    "Too many hash generator requests, to make unlimited requests user api key."
                    "Contact team@herewallet.app to get it for free"
                )
            if r.status != 200:
                raise Exception(
                    f"Error while getting phone hash: {r.status}\n{await r.text()}"
                )
            content = json.loads(await r.text())
            return content["hash"]

    async def get_ft_transfers(self, phone: str) -> List[FtTrustTransaction]:
        """
        Get list of fungible token transfers to phone number
        :param phone: phone number
        :return: list of FtTrustTransaction
        """
        res = (
            await self._account.view_function(
                _PHONE_CONTRACT_ID,
                "get_ft_transfers",
                {"phone": await self._get_phone_hex(phone)},
            )
        ).result
        if res is None:
            return []
        return [FtTrustTransaction(**i) for i in res]

    async def get_near_transfers(self, phone: str) -> List[NearTrustTransaction]:
        """
        Get list of NEAR transfers to phone number
        :param phone: phone number
        :return: list of NEAR transfers
        """
        res = (
            await self._account.view_function(
                _PHONE_CONTRACT_ID,
                "get_transfers",
                {"phone": await self._get_phone_hex(phone)},
            )
        ).result
        if res is None:
            return []
        return [NearTrustTransaction(**i) for i in res]

    async def send_near_to_phone(
        self, phone: str, amount: float, comment: str = "", nowait: bool = False
    ):
        """
        Send NEAR to phone number. Receiver will get sms with link to claim tokens.
        :param phone: +X format phone number
        :param amount: number of NEAR which will be sent
        :param comment: any comment
        :param nowait if True, method will return before transaction is confirmed
        :return: transaction hash ot TransactionResult
        """
        if amount < 0.1:
            raise ValueError("Amount must be >= 0.1 NEAR")
        return await self._account.function_call(
            _PHONE_CONTRACT_ID,
            "send_near_to_phone",
            {"phone": await self._get_phone_hex(phone), "comment": comment},
            amount=int(amount * NEAR),
            nowait=nowait,
            gas=25 * TGAS,
        )

    async def send_ft_to_phone(
        self,
        ft: FtModel,
        phone: str,
        amount: float,
        comment: str = "",
        nowait: bool = False,
    ):
        """
        Send fungible token to phone number. Reciver will get sms with link to claim tokens.
        :param ft: Fungible token model
        :param phone: +X format phone number
        :param amount: number of FT which will be sent
        :param comment:
        :param nowait: if True, method will return before transaction is confirmed
        :return:
        """
        return await self._account.function_call(
            ft.contract_id,
            "ft_transfer_call",
            {
                "msg": await self._get_phone_hex(phone),
                "comment": comment,
                "receiver_id": _PHONE_CONTRACT_ID,
                "amount": str(int(amount * 10**ft.decimal)),
            },
            amount=1,
            gas=50 * TGAS,
            nowait=nowait,
        )

    async def cancel_near_transaction(self, phone: str, index: int):
        """
        Cancel NEAR transfer to phone number. Use index from get_near_transfers() method
        :param phone: phone number
        :param index: index in transaction list
        :return:
        """
        try:
            return await self._account.function_call(
                _PHONE_CONTRACT_ID,
                "cancel_near_transaction",
                {"phone": await self._get_phone_hex(phone), "index": index},
                amount=1,
            )
        except FunctionCallError as e:
            if "`None` value" in str(e.error):
                raise ValueError(f"Transaction with index {index} not found")
            raise e

    async def cancel_ft_transaction(self, phone: str, index: int):
        """
        Cancel fungible token transfer to phone number. Use index from get_ft_transfers() method
        :param phone: phone number
        :param index: index in transaction list
        :return:
        """
        try:
            return await self._account.function_call(
                _PHONE_CONTRACT_ID,
                "cancel_ft_transaction",
                {"phone": await self._get_phone_hex(phone), "index": index},
                amount=1,
            )
        except FunctionCallError as e:
            if "`None` value" in str(e):
                raise ValueError(f"Transaction with index {index} not found")
            raise e



from dataclasses import dataclass


@dataclass
class FtModel:
    contract_id: str
    decimal: int


class FTS:
    USDC = FtModel("dac17f958d2ee523a2206206994597c13d831ec7.factory.bridge.near", 6)
    DAI = FtModel("6b175474e89094c44da98b954eedeac495271d0f.factory.bridge.near", 6)
    USDT = FtModel("a0b86991c6218b36c1d19d4a2e9eb0ce3606eb48.factory.bridge.near", 6)
    wNEAR = FtModel("wrap.near", 24)
    LINEAR = FtModel("linear-protocol.near", 24)
    ETH = FtModel("aurora", 18)
    WBTC = FtModel("2260fac5e5542a773aa44fbcfedf7c193bc2c599.factory.bridge.near", 8)
    stNEAR = FtModel("meta-pool.near", 24)
    META = FtModel("meta-token.near", 24)
    AURORA = FtModel("aaaaaa20d9e0e2461697782ef11675f668207961.factory.bridge.near", 18)
    stHERE = FtModel("storage.herewallet.near", 24)



import base64
import json

import aiohttp
from aiohttp import ClientResponseError, ClientConnectorError
from pynear import constants
from pynear.exceptions.exceptions import RpcNotAvailableError
from pynear.exceptions.provider import (
    UnknownBlockError,
    InvalidAccount,
    NoContractCodeError,
    UnknownAccount,
    TooLargeContractStateError,
    UnavailableShardError,
    NoSyncedBlocksError,
    InternalError,
    NoSyncedYetError,
    InvalidTransactionError,
    RpcTimeoutError,
    UnknownAccessKeyError,
    ERROR_CODE_TO_EXCEPTION,
)


PROVIDER_CODE_TO_EXCEPTION = {
    "UNKNOWN_BLOCK": UnknownBlockError,
    "INVALID_ACCOUNT": InvalidAccount,
    "UNKNOWN_ACCOUNT": UnknownAccount,
    "NO_CONTRACT_CODE": NoContractCodeError,
    "TOO_LARGE_CONTRACT_STATE": TooLargeContractStateError,
    "UNAVAILABLE_SHARD": UnavailableShardError,
    "NO_SYNCED_BLOCKS": NoSyncedBlocksError,
    "INTERNAL_ERROR": InternalError,
    "NOT_SYNCED_YET": NoSyncedYetError,
    "INVALID_TRANSACTION": InvalidTransactionError,
    "TIMEOUT_ERROR": RpcTimeoutError,
    "UNKNOWN_ACCESS_KEY": UnknownAccessKeyError,
}


class JsonProvider(object):
    def __init__(self, rpc_addr):
        if isinstance(rpc_addr, tuple):
            self._rpc_addresses = ["http://{}:{}".format(*rpc_addr)]
        elif isinstance(rpc_addr, list):
            self._rpc_addresses = rpc_addr
        else:
            self._rpc_addresses = [rpc_addr]

    async def call_rpc_request(self, method, params, timeout=60):
        j = {"method": method, "params": params, "id": "dontcare", "jsonrpc": "2.0"}

        content = None
        for rpc_addr in self._rpc_addresses:
            try:
                async with aiohttp.ClientSession() as session:
                    r = await session.post(rpc_addr, json=j, timeout=timeout)
                    r.raise_for_status()
                    content = json.loads(await r.text())
                if self._rpc_addresses[0] != rpc_addr:
                    self._rpc_addresses.remove(rpc_addr)
                    self._rpc_addresses.insert(0, rpc_addr)
                    print("Switching RPC to %s" % rpc_addr)
                break
            except ClientResponseError:
                continue
            except ClientConnectorError:
                continue
            except ConnectionError:
                continue
        return content

    @staticmethod
    def get_error_from_response(content: dict):
        if "error" in content:
            error_code = content["error"].get("cause", {}).get("name", "")
            body = content["error"]["data"]
            error = PROVIDER_CODE_TO_EXCEPTION.get(error_code, InternalError)(
                body, error_json=content["error"]
            )
            while True:
                if not isinstance(body, dict):
                    break
                key, body = list(body.items())[0]
                if key in ERROR_CODE_TO_EXCEPTION:
                    error = ERROR_CODE_TO_EXCEPTION[key](
                        body, error_json=content["error"]
                    )
                else:
                    break
            return error

    async def json_rpc(self, method, params, timeout=60):
        content = await self.call_rpc_request(method, params, timeout)
        if not content:
            raise RpcNotAvailableError("RPC not available")

        error = self.get_error_from_response(content)
        if error:
            raise error
        return content["result"]

    async def send_tx(self, signed_tx: str, timeout: int = constants.TIMEOUT_WAIT_RPC):
        """
        Send a signed transaction to the network and return the hash of the transaction
        :param signed_tx: base64 encoded signed transaction, str.
        :param timeout: rpc request timeout
        :return:
        """
        return await self.json_rpc("broadcast_tx_async", [signed_tx], timeout=timeout)

    async def send_tx_and_wait(
        self, signed_tx: str, timeout: int = constants.TIMEOUT_WAIT_RPC
    ):
        """
        Send a signed transaction to the network and wait for it to be included in a block
        :param signed_tx: base64 encoded signed transaction, str
        :param timeout: rpc request timeout
        :return:
        """
        return await self.json_rpc(
            "broadcast_tx_commit",
            [signed_tx],
            timeout=timeout,
        )

    async def get_status(self):
        async with aiohttp.ClientSession() as session:
            r = await session.get("%s/status" % self._rpc_addresses[0], timeout=60)
            r.raise_for_status()
            return json.loads(await r.text())

    async def get_validators(self):
        return await self.json_rpc("validators", [None])

    async def query(self, query_object):
        return await self.json_rpc("query", query_object)

    async def get_account(self, account_id, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_account",
                "account_id": account_id,
                "finality": finality,
            },
        )

    async def get_access_key_list(self, account_id, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_access_key_list",
                "account_id": account_id,
                "finality": finality,
            },
        )

    async def get_access_key(self, account_id, public_key, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_access_key",
                "account_id": account_id,
                "public_key": public_key,
                "finality": finality,
            },
        )

    async def view_call(self, account_id, method_name, args, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "call_function",
                "account_id": account_id,
                "method_name": method_name,
                "args_base64": base64.b64encode(args).decode("utf8"),
                "finality": finality,
            },
        )

    async def get_block(self, block_id):
        return await self.json_rpc("block", [block_id])

    async def get_chunk(self, chunk_id):
        return await self.json_rpc("chunk", [chunk_id])

    async def get_tx(self, tx_hash, tx_recipient_id):
        return await self.json_rpc("tx", [tx_hash, tx_recipient_id])

    async def get_changes_in_block(self, changes_in_block_request):
        return await self.json_rpc(
            "EXPERIMENTAL_changes_in_block", changes_in_block_request
        )

    async def get_validators_ordered(self, block_hash):
        return await self.json_rpc("EXPERIMENTAL_validators_ordered", [block_hash])

    async def get_light_client_proof(
        self, outcome_type, tx_or_receipt_id, sender_or_receiver_id, light_client_head
    ):
        if outcome_type == "receipt":
            params = {
                "type": "receipt",
                "receipt_id": tx_or_receipt_id,
                "receiver_id": sender_or_receiver_id,
                "light_client_head": light_client_head,
            }
        else:
            params = {
                "type": "transaction",
                "transaction_hash": tx_or_receipt_id,
                "sender_id": sender_or_receiver_id,
                "light_client_head": light_client_head,
            }
        return await self.json_rpc("light_client_proof", params)

    async def get_next_light_client_block(self, last_block_hash):
        return await self.json_rpc("next_light_client_block", [last_block_hash])



from typing import Union, List

import base58
from pyonear.account import (
    AccessKey,
    AccessKeyPermissionFieldless,
    FunctionCallPermission,
)
from pyonear.account_id import AccountId
from pyonear.crypto import ED25519PublicKey, InMemorySigner
from pyonear.crypto_hash import CryptoHash
from pyonear.transaction import (
    CreateAccountAction,
    AddKeyAction,
    DeleteKeyAction,
    TransferAction,
    DeployContractAction,
    FunctionCallAction,
    StakeAction,
    Transaction,
)


def sign_and_serialize_transaction(
    receiver_id, nonce, actions, block_hash, signer: InMemorySigner
) -> str:
    return (
        Transaction(
            signer.account_id,
            signer.public_key,
            nonce,
            AccountId(receiver_id),
            CryptoHash(block_hash),
            actions,
        )
        .sign(signer)
        .to_base64()
    )


def create_create_account_action():
    return CreateAccountAction()


def create_full_access_key_action(pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return AddKeyAction(
        public_key=ED25519PublicKey(pk),
        access_key=AccessKey(0, AccessKeyPermissionFieldless.FullAccess),
    )


def create_function_call_access_key_action(
    pk, allowance: int, receiver_id: str, method_names: List[str]
):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return AddKeyAction(
        ED25519PublicKey(pk),
        AccessKey(0, FunctionCallPermission(receiver_id, method_names, allowance)),
    )


def create_delete_access_key_action(pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    key = ED25519PublicKey(pk)
    return DeleteKeyAction(key)


def create_transfer_action(amount: int):
    return TransferAction(amount)


def create_staking_action(amount, pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    key = ED25519PublicKey(pk)
    return StakeAction(amount, key)


def create_deploy_contract_action(code: bytes):
    return DeployContractAction(code)


def create_function_call_action(method_name: str, args, gas: int, deposit: int):
    return FunctionCallAction(method_name, args, gas, deposit)



from pynear.dapps.core import DappClient
from pynear.dapps.keypom.models import CreateDropModel

_KEYPOM_CONTRACT_ID = "keypom.near"


class KeyPom(DappClient):
    """
    Client to keypom.near contract
    With this contract you can send NEAR and fungible tokens to linkdrop and receive FT/NFT/NEAR from other linkdrops
    """

    def __init__(self, account, contract_id=_KEYPOM_CONTRACT_ID):
        """

        :param account:
        :param contract_id: keypom contract id
        :param network: "mainnet" or "testnet"
        """
        if account.chain_id != "mainnet":
            raise ValueError("Only mainnet is supported")
        super().__init__(account)
        self.contract_id = contract_id

    def create_drop(
        self,
        drop: CreateDropModel,
    ) -> str:
        """

        :param drop: CreateDropModel
        :return: drop id
        """
        res = await self._account.view_function(
            self.contract_id,
            "create_drop",
            drop.dict(),
        )
        return res.result

    async def claim(self, account_id: str, password: str):
        """

        :param account_id: linkdrop receiver account id
        :param password: linkdrop password
        :return:
        """
        return await self._account.view_function(
            self.contract_id,
            "claim",
            {"account_id": account_id, password: password},
        )



import asyncio
import collections
import json
from typing import List, Union, Tuple, Dict, Optional

import base58
from pyonear.account_id import AccountId
from pyonear.crypto import InMemorySigner, ED25519SecretKey, Signer
from pyonear.transaction import Action

from pynear import utils
from pynear import constants
from pynear.dapps.ft.async_client import FT
from pynear.dapps.phone.async_client import Phone
from pynear.dapps.staking.async_client import Staking
from pynear.exceptions.exceptions import (
    AccountAlreadyExistsError,
    AccountDoesNotExistError,
    CreateAccountNotAllowedError,
    ActorNoPermissionError,
    DeleteKeyDoesNotExistError,
    AddKeyAlreadyExistsError,
    DeleteAccountStakingError,
    DeleteAccountHasRentError,
    RentUnpaidError,
    TriesToUnstakeError,
    TriesToStakeError,
    FunctionCallError,
    NewReceiptValidationError,
)
from pynear.models import (
    TransactionResult,
    ViewFunctionResult,
    PublicKey,
    AccountAccessKey,
)
from pynear.providers import JsonProvider
from pynear import transactions


_ERROR_TYPE_TO_EXCEPTION = {
    "AccountAlreadyExists": AccountAlreadyExistsError,
    "AccountDoesNotExist": AccountDoesNotExistError,
    "CreateAccountNotAllowed": CreateAccountNotAllowedError,
    "ActorNoPermission": ActorNoPermissionError,
    "DeleteKeyDoesNotExist": DeleteKeyDoesNotExistError,
    "AddKeyAlreadyExists": AddKeyAlreadyExistsError,
    "DeleteAccountStaking": DeleteAccountStakingError,
    "DeleteAccountHasRent": DeleteAccountHasRentError,
    "RentUnpaid": RentUnpaidError,
    "TriesToUnstake": TriesToUnstakeError,
    "TriesToStake": TriesToStakeError,
    "FunctionCallError": FunctionCallError,
    "NewReceiptValidationError": NewReceiptValidationError,
}


class ViewFunctionError(Exception):
    pass


class Account(object):
    """
    This class implement all blockchain functions for your account
    """

    _access_key: dict
    _lock: asyncio.Lock = None
    _lock_by_pk: Dict[str, asyncio.Lock] = {}
    _latest_block_hash: str
    _latest_block_hash_ts: float = 0
    chain_id: str = "mainnet"

    def __init__(
        self,
        account_id: str = None,
        private_key: Union[List[Union[str, bytes]], str, bytes] = None,
        rpc_addr="https://rpc.mainnet.near.org",
    ):
        self._provider = JsonProvider(rpc_addr)
        self.account_id = account_id
        if private_key is None:
            private_keys = []
        elif isinstance(private_key, list):
            private_keys = private_key
        elif isinstance(private_key, str):
            private_keys = [private_key]
        elif isinstance(private_key, bytes):
            private_keys = [private_key]
        else:
            return

        self._free_signers = asyncio.Queue()
        self._signers = []
        for pk in private_keys:
            if isinstance(pk, str):
                pk = base58.b58decode(pk.replace("ed25519:", ""))
            key = ED25519SecretKey(pk)
            singer = InMemorySigner(
                AccountId(account_id),
                key.public_key(),
                key,
            )
            self._free_signers.put_nowait(singer)
            self._signers.append(singer)

    async def startup(self):
        """
        Initialize async object
        :return:
        """
        self._lock = asyncio.Lock()
        self._lock_by_pk = collections.defaultdict(asyncio.Lock)
        self.chain_id = (await self._provider.get_status())["chain_id"]

    async def _update_last_block_hash(self):
        """
        Update last block hash& If it's older than 50 block before, transaction will fail
        :return: last block hash
        """
        if self._latest_block_hash_ts + 50 > utils.timestamp():
            return
        self._latest_block_hash = (await self._provider.get_status())["sync_info"][
            "latest_block_hash"
        ]
        self._latest_block_hash_ts = utils.timestamp()

    async def _sign_and_submit_tx(
        self, receiver_id, actions: List[Action], nowait=False
    ) -> Union[TransactionResult, str]:
        """
        Sign transaction and send it to blockchain
        :param receiver_id:
        :param actions: list of actions
        :param nowait: if nowait is True, return transaction hash, else wait execution
        confirm and return TransactionResult
        :return: transaction hash or TransactionResult
        """
        if not self._signers:
            raise ValueError("You must provide a private key or seed to call methods")
        signer = await self._free_signers.get()

        try:
            access_key = await self.get_access_key(signer)
            await self._update_last_block_hash()

            block_hash = base58.b58decode(self._latest_block_hash.encode("utf8"))
            serialized_tx = transactions.sign_and_serialize_transaction(
                receiver_id,
                access_key.nonce + 1,
                actions,
                block_hash,
                signer,
            )
            if nowait:
                return await self._provider.send_tx(serialized_tx)
            result = await self._provider.send_tx_and_wait(serialized_tx)

            if "Failure" in result["status"]:
                error_type, args = list(
                    result["status"]["Failure"]["ActionError"]["kind"].items()
                )[0]
                raise _ERROR_TYPE_TO_EXCEPTION[error_type](**args)
            return TransactionResult(**result)
        except Exception as e:
            raise e
        finally:
            print("Unlock sender")
            await self._free_signers.put(signer)

    @property
    def signer(self) -> Optional[InMemorySigner]:
        if not self._signers:
            return None
        return self._signers[0]

    @property
    def provider(self) -> JsonProvider:
        return self._provider

    async def get_access_key(self, signer: Signer = None) -> AccountAccessKey:
        """
        Get access key for current account
        :return: AccountAccessKey
        """
        if signer is None:
            signer = self._signers[0]
        resp = await self._provider.get_access_key(
            self.account_id, str(signer.public_key)
        )
        if "error" in resp:
            raise ValueError(resp["error"])
        return AccountAccessKey(**resp)

    async def get_access_key_list(self, account_id: str = None) -> List[PublicKey]:
        """
        Get access key list for account_id, if account_id is None, get access key list for current account
        :param account_id:
        :return: list of PublicKey
        """
        if account_id is None:
            account_id = self.account_id
        resp = await self._provider.get_access_key_list(account_id)
        result = []
        if "keys" in resp and isinstance(resp["keys"], list):
            for key in resp["keys"]:
                result.append(PublicKey(**key))
        return result

    async def fetch_state(self) -> dict:
        """Fetch state for given account."""
        return await self._provider.get_account(self.account_id)

    async def send_money(
        self, account_id: str, amount: int, nowait: bool = False
    ) -> TransactionResult:
        """
        Send money to account_id
        :param account_id: receiver account id
        :param amount: amount in yoctoNEAR
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        return await self._sign_and_submit_tx(
            account_id, [transactions.create_transfer_action(amount)], nowait
        )

    async def function_call(
        self,
        contract_id: str,
        method_name: str,
        args: dict,
        gas: int = constants.DEFAULT_ATTACHED_GAS,
        amount: int = 0,
        nowait: bool = False,
    ):
        """
        Call function on smart contract
        :param contract_id: smart contract address
        :param method_name: call method name
        :param args: json params for method
        :param gas: amount of attachment gas. Default is 200000000000000
        :param amount: amount of attachment NEAR, Default is 0
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        args = json.dumps(args).encode("utf8")
        return await self._sign_and_submit_tx(
            contract_id,
            [transactions.create_function_call_action(method_name, args, gas, amount)],
            nowait,
        )

    async def create_account(
        self,
        account_id: str,
        public_key: Union[str, bytes],
        initial_balance: int,
        nowait=False,
    ):
        """
        Create new account in subdomain of current account. For example, if current account is "test.near",
        you can create "wwww.test.near"
        :param account_id: new account id
        :param public_key: add public key to new account
        :param initial_balance: amount to transfer NEAR to new account
        :param nowait: is nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        actions = [
            transactions.create_create_account_action(),
            transactions.create_full_access_key_action(public_key),
            transactions.create_transfer_action(initial_balance),
        ]
        return await self._sign_and_submit_tx(account_id, actions, nowait)

    async def add_public_key(
        self,
        public_key: Union[str, bytes],
        receiver_id: str,
        method_names: List[str] = None,
        allowance: int = constants.ALLOWANCE,
        nowait=False,
    ):
        """
        Add public key to account with access to smart contract methods
        :param public_key: public_key to add
        :param receiver_id: smart contract account id
        :param method_names: list of method names to allow
        :param allowance: maximum amount of gas to use for this key
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        if method_names is None:
            method_names = []
        actions = [
            transactions.create_function_call_access_key_action(
                public_key, allowance, receiver_id, method_names
            ),
        ]
        return await self._sign_and_submit_tx(self.account_id, actions, nowait)

    async def add_full_access_public_key(
        self, public_key: Union[str, bytes], nowait=False
    ) -> TransactionResult:
        """
        Add public key to account with full access
        :param public_key: public_key to add
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        actions = [
            transactions.create_full_access_key_action(public_key),
        ]
        return await self._sign_and_submit_tx(self.account_id, actions, nowait)

    async def delete_public_key(self, public_key: Union[str, bytes], nowait=False):
        """
        Delete public key from account
        :param public_key: public_key to delete
        :param nowait: is nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        actions = [
            transactions.create_delete_access_key_action(public_key),
        ]
        return await self._sign_and_submit_tx(self.account_id, actions, nowait)

    async def deploy_contract(self, contract_code: bytes, nowait=False):
        """
        Deploy smart contract to account
        :param contract_code: smart contract code
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        return await self._sign_and_submit_tx(
            self.account_id,
            [transactions.create_deploy_contract_action(contract_code)],
            nowait,
        )

    async def stake(self, public_key: str, amount: str, nowait=False):
        """
        Stake NEAR on account. Account must have enough balance to be in validators pool
        :param public_key: public_key to stake
        :param amount: amount of NEAR to stake
        :param nowait: if nowait is True, return transaction hash, else wait execution
        :return: transaction hash or TransactionResult
        """
        return await self._sign_and_submit_tx(
            self.account_id,
            [transactions.create_staking_action(public_key, amount)],
            nowait,
        )

    async def view_function(
        self, contract_id: str, method_name: str, args: dict
    ) -> ViewFunctionResult:
        """
        Call view function on smart contract. View function is read only function, it can't change state
        :param contract_id: smart contract account id
        :param method_name: method name to call
        :param args: json args to call method
        :return: result of view function call
        """
        result = await self._provider.view_call(
            contract_id, method_name, json.dumps(args).encode("utf8")
        )
        if "error" in result:
            raise ViewFunctionError(result["error"])
        result["result"] = json.loads("".join([chr(x) for x in result["result"]]))
        return ViewFunctionResult(**result)

    async def get_balance(self, account_id: str = None) -> int:
        """
        Get account balance
        :param account_id: if account_id is None, return balance of current account
        :return: balance of account in yoctoNEAR
        """
        if account_id is None:
            account_id = self.account_id
        return int((await self._provider.get_account(account_id))["amount"])

    @property
    def phone(self):
        """
        Get client for phone.herewallet.near
        :return: Phone(self)
        """
        return Phone(self)

    @property
    def ft(self):
        """
        Get client for fungible tokens
        :return: FT(self)
        """
        return FT(self)

    @property
    def staking(self):
        """
        Get client for staking
        :return: Staking(self)
        """
        return Staking(self)



import json
from dataclasses import dataclass


class RpcNotAvailableError(Exception):
    pass


class ActionErrorKind(Exception):
    pass


@dataclass
class AccountAlreadyExistsError(ActionErrorKind):
    """
    Happens when CreateAccount action tries to create an account with account_id which is already exists in the storage
    """

    account_id: str


class AccountDoesNotExistError(ActionErrorKind):
    """
    Happens when TX receiver_id doesn't exist (but action is not Action::CreateAccount)
    """

    pass


class CreateAccountNotAllowedError(ActionErrorKind):
    """
    A newly created account must be under a namespace of the creator account
    """

    account_id: str
    predecessor_id: str

    def __init__(self, account_id, predecessor_id):
        self.predecessor_id = predecessor_id
        self.account_id = account_id


class ActorNoPermissionError(ActionErrorKind):
    """
    Administrative actions like `DeployContract`, `Stake`, `AddKey`, `DeleteKey`.
    can be proceed only if sender=receiver or the first TX action is a `CreateAccount` action
    """

    account_id: str
    actor_id: str

    def __init__(self, account_id, actor_id):
        self.actor_id = actor_id
        self.account_id = account_id


class DeleteKeyDoesNotExistError(ActionErrorKind):
    """
    Account tries to remove an access key that doesn't exist
    """

    account_id: str
    public_key: str

    def __init__(self, account_id, public_key):
        self.actor_id = public_key
        self.account_id = account_id


class AddKeyAlreadyExistsError(ActionErrorKind):
    """
    The public key is already used for an existing access key
    """

    account_id: str
    public_key: str

    def __init__(self, account_id, public_key):
        self.actor_id = public_key
        self.account_id = account_id


class DeleteAccountStakingError(ActionErrorKind):
    """
    Account is staking and can not be deleted
    """

    account_id: str

    def __init__(self, account_id):
        self.account_id = account_id


class DeleteAccountHasRentError(ActionErrorKind):
    """
    Foreign sender (sender=!receiver) can delete an account only if a target account hasn't enough tokens to pay rent
    """

    account_id: str
    balance: str

    def __init__(self, account_id, balance):
        self.balance = balance
        self.account_id = account_id


class RentUnpaidError(ActionErrorKind):
    """
    ActionReceipt can't be completed, because the remaining balance will not be enough to pay rent.
    """

    account_id: str
    amount: str

    def __init__(self, account_id, amount):
        self.amount = amount
        self.account_id = account_id


class TriesToUnstakeError(ActionErrorKind):
    """
    Account is not yet staked, but tries to unstake
    """

    account_id: str

    def __init__(self, account_id):
        self.account_id = account_id


class TriesToStakeError(ActionErrorKind):
    """
    The account doesn't have enough balance to increase the stake.
    """

    account_id: str
    stake: str
    locked: str
    balance: str

    def __init__(self, account_id, stake, locked, balance):
        self.account_id = account_id
        self.stake = stake
        self.locked = locked
        self.balance = balance


class FunctionCallError(ActionErrorKind):
    """
    An error occurred during a `FunctionCall` Action.
    """

    error: dict

    def __init__(self, **kwargs):
        super().__init__(json.dumps(kwargs))
        self.error = kwargs


class NewReceiptValidationError(ActionErrorKind):
    """
    Error occurs when a new `ActionReceipt` created by the `FunctionCall` action fails
    """

    pass



from dataclasses import dataclass
from enum import Enum
from typing import List, Any, Optional, Union


class ReceiptOutcome:
    logs: List[str]
    metadata: dict
    receipt_ids: List[str]
    status: dict
    tokens_burnt: str
    executor_id: str
    gas_burnt: int

    def __init__(self, data):
        self.logs = data["outcome"]["logs"]
        self.metadata = data["outcome"]["metadata"]
        self.receipt_ids = data["outcome"]["receipt_ids"]
        self.status = data["outcome"]["status"]
        self.tokens_burnt = data["outcome"]["tokens_burnt"]
        self.gas_burnt = data["outcome"]["gas_burnt"]


class TransactionData:
    hash: str
    public_key: str
    receiver_id: str
    signature: str
    signer_id: str
    nonce: int
    actions: List[dict]

    def __init__(
        self,
        hash,
        public_key,
        receiver_id,
        signature,
        signer_id,
        nonce,
        actions,
        **kargs,
    ):
        self.actions = actions
        self.nonce = nonce
        self.signer_id = signer_id
        self.public_key = public_key
        self.receiver_id = receiver_id
        self.signature = signature
        self.hash = hash

    @property
    def url(self):
        return f"https://explorer.near.org/transactions/{self.hash}"


class TransactionResult:
    receipt_outcome: List[ReceiptOutcome]
    transaction_outcome: ReceiptOutcome
    status: dict
    transaction: TransactionData

    def __init__(self, receipts_outcome, transaction_outcome, transaction, status):
        self.status = status
        self.transaction = TransactionData(**transaction)
        self.transaction_outcome = ReceiptOutcome(transaction_outcome)

        self.receipt_outcome = []
        for ro in receipts_outcome:
            self.receipt_outcome.append(ReceiptOutcome(ro))

    @property
    def logs(self):
        logs = self.transaction_outcome.logs
        for ro in self.receipt_outcome:
            logs.extend(ro.logs)
        return logs


class ViewFunctionResult:
    block_hash: str
    block_height: str
    logs: List[str]
    result: Any

    def __init__(self, block_hash, block_height, logs, result):
        self.block_hash = block_hash
        self.block_height = block_height
        self.logs = logs
        self.result = result


class PublicKeyPermissionType(str, Enum):
    FULL_ACCESS = "FullAccess"
    FUNCTION_CALL = "FunctionCall"


@dataclass
class AccessKey:
    permission_type: PublicKeyPermissionType
    nonce: int
    allowance: Optional[str] = None
    receiver_id: Optional[str] = None
    method_names: Optional[List[str]] = None

    @classmethod
    def build(cls, data: dict) -> "AccessKey":
        if data["permission"] == PublicKeyPermissionType.FULL_ACCESS:
            return cls(nonce=data["nonce"], permission_type=PublicKeyPermissionType.FULL_ACCESS)

        permission_type, permission_data = list(data["permission"].items())[0]
        return cls(
            nonce=data["nonce"],
            permission_type=PublicKeyPermissionType.FUNCTION_CALL,
            **permission_data,
        )


@dataclass
class PublicKey:
    public_key: str
    access_key: AccessKey

    @classmethod
    def build(cls, data: dict) -> "PublicKey":
        return cls(
            data["public_key"],
            AccessKey.build(data["access_key"]),
        )


@dataclass
class AccountAccessKey:
    block_hash: str
    block_height: int
    nonce: int
    permission: Union[str, dict]



from typing import Optional

from pynear.dapps.core import DappClient
from pynear.dapps.staking.exceptions import NotEnoughBalance
from pynear.dapps.staking.models import StakingData
from pynear.exceptions.exceptions import FunctionCallError

CONTRACT_ID = {
    "mainnet": "storage.herewallet.near",
    "testnet": "storage.herewallet.testnet",
}


class Staking(DappClient):
    async def transfer(
        self,
        receiver_id: str,
        amount: float,
        memo: str = "",
        nowait: bool = False,
    ):
        """
        Transfer hNEAR to account

        :param receiver_id: receiver account id
        :param amount: amount in yoctoNEAR
        :param memo: comment
        :param nowait if True, method will return before transaction is confirmed
        :return: transaction hash ot TransactionResult
        """

        try:
            return await self._account.function_call(
                CONTRACT_ID[self._account.chain_id],
                "ft_transfer",
                {
                    "receiver_id": receiver_id,
                    "amount": str(amount),
                    "msg": memo,
                },
                amount=1,
                nowait=nowait,
            )
        except FunctionCallError as e:
            if "The account doesn't have enough balance" in e.error["ExecutionError"]:
                raise NotEnoughBalance(e)
            raise e

    async def transfer_call(
        self,
        receiver_id: str,
        amount: int,
        memo: str = "",
        nowait: bool = False,
    ):
        """
        Transfer hNEAR to account and call ft_on_transfer() method in receiver contract

        :param receiver_id: receiver account id
        :param amount: amount in yoctoNEAR
        :param memo: comment
        :param nowait if True, method will return before transaction is confirmed
        :return: transaction hash ot TransactionResult
        """
        return await self._account.function_call(
            CONTRACT_ID[self._account.chain_id],
            "ft_transfer_call",
            {
                "receiver_id": receiver_id,
                "amount": str(amount),
                "msg": memo,
            },
            amount=1,
            nowait=nowait,
        )

    async def get_staking_amount(self, account_id: str = None) -> int:
        """
        Get staking balance of account.

        :param account_id: account id
        :param nowait if True, method will return before transaction is confirmed
        :return: int balance in yoctoNEAR
        """
        if account_id is None:
            account_id = self._account.account_id
        res = (
            await self._account.view_function(
                CONTRACT_ID[self._account.chain_id],
                "ft_balance_of",
                {"account_id": account_id},
            )
        ).result
        if res:
            return int(res)
        return 0

    async def get_user(self, account_id: str = None) -> Optional[StakingData]:
        """
        Get user staking parameters

        :param account_id: account id
        :return: StakingData
        """
        if account_id is None:
            account_id = self._account.account_id
        res = (
            await self._account.view_function(
                CONTRACT_ID[self._account.chain_id],
                "get_user",
                {"account_id": account_id},
            )
        ).result
        if res:
            return StakingData(**res)

    async def stake(self, amount: int, nowait: bool = False):
        """
        Deposit staking for account

        :param amount: in amount of yoctoNEAR
        :param nowait: if True, method will return before transaction is confirmed
        :return: transaction hash or TransactionResult
        """
        return await self._account.function_call(
            CONTRACT_ID[self._account.chain_id],
            "storage_deposit",
            {},
            amount=amount,
            nowait=nowait,
        )

    async def unstake(self, amount: int, nowait: bool = False):
        """
        Withdraw from staking

        :param amount: in amount of yoctoNEAR
        :param nowait: if True, method will return before transaction is confirmed
        :return: transaction hash or TransactionResult
        """
        try:
            return await self._account.function_call(
                CONTRACT_ID[self._account.chain_id],
                "storage_withdraw",
                {"amount": str(int(amount))},
                amount=1,
                nowait=nowait,
            )
        except FunctionCallError as e:
            if "The account doesn't have enough balance" in e:
                raise NotEnoughBalance
            raise e

    async def receive_dividends(self, nowait=False):
        """
        Receive dividends

        :param nowait: if True, method will return before transaction is confirmed
        :return: transaction hash ot TransactionResult
        """
        return await self._account.function_call(
            CONTRACT_ID[self._account.chain_id],
            "receive_dividends",
            {},
            amount=1,
            nowait=nowait,
        )



from enum import Enum
from typing import Optional, Dict, List

from pydantic import BaseModel


class DropPermissionEnum(str, Enum):
    CLAIM = "claim"
    CREATE_ACCOUNT_AND_CLAIM = "create_account_and_claim"


class DropKeyConfig(BaseModel):
    remaining_uses: int
    last_used: int
    allowance: int
    key_id: int
    pw_per_use: Optional[Dict[int, bytes]]
    pw_per_key: Optional[bytes]


class DropTimeConfig(BaseModel):
    start: Optional[int]
    end: Optional[int]
    throttle: Optional[int]
    interval: Optional[int]


class DropUsageConfig(BaseModel):
    permissions: Optional[DropPermissionEnum]
    refund_deposit: Optional[bool]
    auto_delete_drop: Optional[bool]
    auto_withdraw: Optional[bool]


class DropConfig(BaseModel):
    uses_per_key: Optional[int]
    time: Optional[DropTimeConfig]
    usage: Optional[DropUsageConfig]
    root_account_id: Optional[str]


class SimpleData(BaseModel):
    lazy_register: Optional[bool]


class JsonFTData(BaseModel):
    contract_id: str
    sender_id: str
    balance_per_use: int


class JsonNFTData(BaseModel):
    sender_id: str
    contract_id: str


class FCConfig(BaseModel):
    attached_gas: Optional[int]


class MethodData(BaseModel):
    receiver_id: str
    method_name: str
    args: str
    attached_deposit: int
    account_id_field: Optional[str]
    drop_id_field: Optional[str]
    key_id_field: Optional[str]


class FCData(BaseModel):
    methods: List[Optional[List[MethodData]]]
    config: Optional[FCConfig]


class JsonKeyInfo(BaseModel):
    drop_id: str
    pk: str
    cur_key_use: int
    remaining_uses: int
    last_used: int
    allowance: int
    key_id: int


class JsonPasswordForUse(BaseModel):
    pw: str
    key_use: int


class CreateDropModel(BaseModel):
    public_keys: Optional[List[str]]
    deposit_per_use: int
    drop_id: Optional[str]
    config: Optional[DropConfig]
    metadata: Optional[str]
    simple: Optional[SimpleData]
    ft: Optional[JsonFTData]
    nft: Optional[JsonNFTData]
    fc: Optional[FCData]
    passwords_per_use: Optional[List[Optional[List[JsonPasswordForUse]]]]
    passwords_per_key: Optional[List[Optional[str]]]
